import java.util.*;

/**
 * Esercitazione LOO n.5 - A.A.2005/2006
 * Gestisce un parcheggio che deve interagire 
 * con i sistemi di emissione e lettura dei tagliandi di sosta e 
 * di gettoniera.
 * 
 * @author Nicola Asuni, Fabrizio Solinas
 * @version 2005-12-10
 */
public class ParkingControl {
    /**
     * Il numero di posti disponibili.
     */
    private int places;
    
    /**
     * La lista dei biglietti attivi, ovvero relativi a macchine parcheggiate.
     */
    private ArrayList<Ticket> tickets;
    
    /**
     * Il progressivo dell'ultimo biglietto emesso.
     */
    private long lastTicket;
    
    /**
     * Tempo minimo accettabile tra il momento del pagamento e quello dell'uscita in millisecondi.
     */
    private static long PAYMENTMINTIME = 10L * 60L * 1000L;
    
    /**
     * Tempo massimo accettabile in millisecondi di differenza tra il tempo di ingresso
     * registrato e quello letto sul biglietto.
     */
    private static long EXITMINTIME = 2L * 60L * 1000L;
    
    /**
     * numero di  millisecondi in un'ora
     */
    private static long HOURMILLISEC = 60L * 60L * 1000L;

    /**
     * Costruttore della classe che si occupa di inizializzare lo stato 
     * dell'oggetto, il parametro individua il numero di posti disponibili.
     * 
     * @param posti numero di posti disponibili nel parcheggio.
     */
    public ParkingControl(int posti) {
        if (posti > 0) {
            places = posti;
        } else {
            places = 1;
        }
        lastTicket = 0;
        tickets = new ArrayList<Ticket>();   
    }
    
    /**
     * Invocato dalla macchina emettitrice quando un'auto entra nel parcheggio, 
     * se il valore di ritorno è positivo è il numero di serie del ticket emesso 
     * da registrare sulla banda magnetica e indica che c'è posto nel parcheggio, 
     * se il valore è 0 indica che non c'è posto nel parcheggio.
     * 
     * @return numero di serie del biglietto se il posto è disponibile, zero altrimenti.
     */
    public long addCar() {
        if (places > 0) {
           lastTicket++;
           places--;
           tickets.add(new Ticket(lastTicket));
           return(lastTicket);
        } 
        return(0);
    }
    
    /**
     * Dà l'autorizzazione al sollevamento della sbarra quando il lettore all'uscita 
     * del parcheggio legge il biglietto con il numero di serie e il tempo di ingresso 
     * passati come parametri, devono essere eseguiti i controlli per verificare che 
     * il pagamento sia stato effettuato nei 10 minuti precedenti e che il tempo di 
     * ingresso sia coincidente (con una tolleranza di 2 minuti) con quello registrato 
     * nel sistema.
     * 
     * @param ticketnum numero di serie del biglietto.
     * @param Date time tempo (timestamp) di ingresso.
     * @return true se l'uscita � autorizzata, false altrimenti.
     */
    public boolean exit(long ticketnum, Date time) {
        Date currenttime = new Date();
        int ticketid = getTicket(ticketnum);
        if (ticketid >= 0) {
            Ticket currentticket = (Ticket)tickets.get(ticketid);
            if (currentticket.getTimeOut() != null) {
                if (((currenttime.getTime() - currentticket.getTimeOut().getTime()) <= PAYMENTMINTIME) 
                    && (Math.abs(time.getTime() - currentticket.getTimeIn().getTime()) <= EXITMINTIME)){
                    places++;
                    return true;
                }
            }
        }
        return false;
    }
    
    /**
     * Restituisce l'indice dell'oggetto Ticket dell'Array che ha il 
     * numero di serie specificato.
     * @author Nicola Asuni, Fabrizio Solinas
     * 
     * @param ticketnum numero di serie del biglietto. 
     * @return indice dell'oggetto nell'ArrayList, -1 se non esiste.
     */
    private int getTicket(long ticketnum) {
        // troviamo l'oggetto ticket che ci interessa
        Iterator i = tickets.iterator();
        while(i.hasNext()) {
            Ticket currentticket = (Ticket)i.next();
            if (currentticket.getSerialNumber() == ticketnum)  {
                return tickets.indexOf(currentticket);
            }
        }
        return -1;
    }
    
    /**
     * Invocato dalla gettoniera per ottenere il costo della sosta sulla base del numero 
     * di serie del biglietto, il parcheggio costa 1 euro all'ora dalle 8 del mattino 
     * alle 13, 50 centesimi all'ora dalle 13 alle 17 e 1 euro all'ora dalle 17 alle 20, 
     * dalle 20 alle 8 il parcheggio � chiuso.
     * 
     * @param ticketnum numero di serie del biglietto.
     * @return costo della sosta, 0 in caso di errore.
     */
    public int getPrice(long ticketnum) {
        int ticketid = getTicket(ticketnum);
        int cost = 0;
        if (ticketid >= 0) {
            Ticket currentticket = (Ticket)tickets.get(ticketid);
            // supponiamo che un parcheggio non possa protrarsi per pi� un giorno
            
            Calendar in = new GregorianCalendar(); 
            in.setTime(currentticket.getTimeIn()); // tempo di ingresso
            
            Calendar out = new GregorianCalendar();
            out.setTime(currentticket.getTimeOut()); // tempo di uscita
            // si poteva fare una lista di oggetti che contenevano i dati degli 
            // intervalli temporali e le relative tariffe e poi fare un ciclo...
            // ai fini del testing ci basta un po' di hardcoding :)
            cost += getIntervalPrice(100, 8, 13, in, out);
            cost += getIntervalPrice(50, 13, 17, in, out);
            cost += getIntervalPrice(100, 17, 20, in, out);
            
            tickets.get(getTicket(ticketnum)).setCost(cost);
            return cost;
        }
        return 0;
    }
    
    /**
     * Restituisce il costo per l'intervallo specificato.
     * @author Nicola Asuni, Fabrizio Solinas
     * 
     * @param int costo costo orario dell'intervallo
     * @param hourstart ora di inizio dell'intervallo (compresa)
     * @param hourend ora di fine dell'intervallo (non compresa)
     * @param Calendar in data di ingresso
     * @param Calendar out data di uscita
     * @return costo in centesimi di euro per l'occupazione della fascia oraria specificata
     */
    public int getIntervalPrice(int costo, int hourstart, int hourend, Calendar in, Calendar out) {
        return Math.round(((float)costo / (float)HOURMILLISEC) * getIntervalTime(hourstart, hourend, in, out));
    }
    
    /**
     * Restituisce il numero di minuti di occupazione per l'intervallo specificato.
     * @author Nicola Asuni, Fabrizio Solinas
     * 
     * @param hourstart ora di inizio dell'intervallo (compresa)
     * @param hourend ora di fine dell'intervallo (non compresa)
     * @param Calendar inx data di ingresso
     * @param Calendar outx data di uscita
     * @return numero di minuti di occupazione
     */
    public long getIntervalTime(int hourstart, int hourend, Calendar inx, Calendar outx) {
        Calendar in = (Calendar)inx.clone();
        Calendar out = (Calendar)outx.clone();
        // controlla se l'intervallo � esterno
        if ((out.get(Calendar.HOUR_OF_DAY) < hourstart) || (in.get(Calendar.HOUR_OF_DAY) >= hourend))  {
            return 0;
        }
        if (in.get(Calendar.HOUR_OF_DAY) < hourstart) {
                in.set(Calendar.HOUR_OF_DAY, hourstart);
                in.set(Calendar.MINUTE, 0);
        }
        if (out.get(Calendar.HOUR_OF_DAY) >= hourend) {
                out.set(Calendar.HOUR_OF_DAY, hourend);
                out.set(Calendar.MINUTE, 0);
        }
        return Math.round(out.getTimeInMillis() - in.getTimeInMillis());
    }
    
    /**
     * Invocato dalla gettoniera per comunicare l'avvenuto pagamento della sosta del 
     * biglietto con il numero di serie indicato.
     * 
     * @param ticketnum numero di serie del biglietto. 
     */
    public void payment(long ticketnum) {
        // imposto la data di pagamento
        tickets.get(getTicket(ticketnum)).setTimeOut(new Date());
    }
    
    /**
     * Restituisce il numero di macchine presenti nel parcheggio.
     * 
     * @return numero di macchine presenti nel parcheggio.
     */
    public int howMany() {
        int numcars = 0;
        Date currenttime = new Date();
        // troviamo l'oggetto ticket che ci interessa
        Iterator i = tickets.iterator();
        while(i.hasNext()) {
            Ticket currentticket = (Ticket)i.next();
            if (currentticket.getTimeOut() == null)  {
                numcars++;
            }
        }
        return numcars;
    }
    
    /**
     * Restituisce l'importo, in centesimi di euro, medio giornaliero incassato 
     * nell'ultimo mese.
     * 
     * @return importo, in centesimi di euro, medio giornaliero incassato nell'ultimo mese.
     */
    public int averageRevenue() {
        Calendar curdate = new GregorianCalendar();
        curdate.add(Calendar.MONTH, -1);
        int refyear = curdate.get(Calendar.YEAR);
        int refmonth = curdate.get(Calendar.MONTH);
        int monthdays = curdate.getMaximum(Calendar.DAY_OF_MONTH);
        int totalcost = 0;
        
        Iterator i = tickets.iterator();
        while(i.hasNext()) {
            Ticket currentticket = (Ticket)i.next();
            curdate.setTime(currentticket.getTimeIn());
            if ((curdate.get(Calendar.YEAR) == refyear) 
                && (curdate.get(Calendar.MONTH) == refmonth))  {
                totalcost += currentticket.getCost();
            }
        }
        if (monthdays > 0) {
            return Math.round((float)totalcost / (float)monthdays);
        } else {
            return 0;
        }
    }
    
    /**
     * Restituisce l'occupazione media del parcheggio arrotondata all'intero pi� vicino, 
     * del giorno precedente, per ogni ora del giorno; in altre parole il primo elemento 
     * dell'array indicher� il numero medio di macchine presenti dalle 8 alle 9 e cos� 
     * via sino all'ultima ora dalle 19 alle 20; il calcolo della media � inteso come 
     * numero di minuti totale diviso 60.
     * 
     * @return Occupazione media del parcheggio.
     */
    public int[] averageOccupancy() {
        Calendar indate = new GregorianCalendar(); // tempo di inizio occupazione
        Calendar outdate = new GregorianCalendar(); // tempo fine occupazione
        indate.add(Calendar.DAY_OF_MONTH, - 1); // prendo il giorno precedente
        int refyear = indate.get(Calendar.YEAR);
        int refmonth = indate.get(Calendar.MONTH);
        int refday = indate.get(Calendar.DAY_OF_MONTH);
        int hour; // hour iterator
        long[] hoursum = new long[12]; // sommma dei millisecondi di occupazione per ogni ora
        int[] houraverage = new int[12]; // array da restituire
        
        Iterator i = tickets.iterator();
        while(i.hasNext()) {
            Ticket currentticket = (Ticket)i.next();
            indate.setTime(currentticket.getTimeIn());

            if (currentticket.getTimeOut() != null)  {
                outdate.setTime(currentticket.getTimeOut());
                if ((indate.get(Calendar.YEAR) == refyear) 
                && (indate.get(Calendar.MONTH) == refmonth)
                && (indate.get(Calendar.DAY_OF_MONTH) == refday)) {
                    for (hour = Math.max(indate.get(Calendar.HOUR_OF_DAY),8); (hour < 20) && (hour <= outdate.get(Calendar.HOUR_OF_DAY)); hour++) {
                        hoursum[hour-8] += getIntervalTime(hour, hour + 1, indate, outdate);
                    }
                }
            }
        }
        // calcolo la media
        for (hour=0; hour<12; hour++) {
            houraverage[hour] = Math.round((float)hoursum[hour] / (float)HOURMILLISEC);
        }
        return houraverage;
    }
}
